from django.contrib.auth.models import User
from django.db import models
from django.test import TestCase
from django.test.utils import override_settings

from guardian.shortcuts import assign_perm

from django_fsm import FSMField, transition, has_transition_perm


class ObjectPermissionTestModel(models.Model):
    state = FSMField(default="new")

    @transition(field=state, source='new', target='published',
                on_error='failed', permission='testapp.can_publish_objectpermissiontestmodel')
    def publish(self):
        pass

    class Meta:
        app_label = 'testapp'

        permissions = [
            ('can_publish_objectpermissiontestmodel', 'Can publish ObjectPermissionTestModel'),
        ]


@override_settings(
    AUTHENTICATION_BACKENDS=('django.contrib.auth.backends.ModelBackend',
                             'guardian.backends.ObjectPermissionBackend'))
class ObjectPermissionFSMFieldTest(TestCase):
    def setUp(self):
        super(ObjectPermissionFSMFieldTest, self).setUp()
        self.model = ObjectPermissionTestModel.objects.create()

        self.unprivileged = User.objects.create(username='unpriviledged')
        self.privileged = User.objects.create(username='object_only_privileged')
        assign_perm('can_publish_objectpermissiontestmodel', self.privileged, self.model)

    def test_object_only_access_success(self):
        self.assertTrue(has_transition_perm(self.model.publish, self.privileged))
        self.model.publish()

    def test_object_only_other_access_prohibited(self):
        self.assertFalse(has_transition_perm(self.model.publish, self.unprivileged))



from django.db import models
from django_fsm import FSMField, FSMKeyField, transition


class Application(models.Model):
    """
    Student application need to be approved by dept chair and dean.
    Test workflow
    """
    state = FSMField(default='new')

    @transition(field=state, source='new', target='draft')
    def draft(self):
        pass

    @transition(field=state, source=['new', 'draft'], target='dept')
    def to_approvement(self):
        pass

    @transition(field=state, source='dept', target='dean')
    def dept_approved(self):
        pass

    @transition(field=state, source='dept', target='new')
    def dept_rejected(self):
        pass

    @transition(field=state, source='dean', target='done')
    def dean_approved(self):
        pass

    @transition(field=state, source='dean', target='dept')
    def dean_rejected(self):
        pass


class FKApplication(models.Model):
    """
    Student application need to be approved by dept chair and dean.
    Test workflow for FSMKeyField
    """
    state = FSMKeyField('testapp.DbState', default='new', on_delete=models.CASCADE)

    @transition(field=state, source='new', target='draft')
    def draft(self):
        pass

    @transition(field=state, source=['new', 'draft'], target='dept')
    def to_approvement(self):
        pass

    @transition(field=state, source='dept', target='dean')
    def dept_approved(self):
        pass

    @transition(field=state, source='dept', target='new')
    def dept_rejected(self):
        pass

    @transition(field=state, source='dean', target='done')
    def dean_approved(self):
        pass

    @transition(field=state, source='dean', target='dept')
    def dean_rejected(self):
        pass


class DbState(models.Model):
    '''
    States in DB
    '''
    id = models.CharField(primary_key=True, max_length=50)

    label = models.CharField(max_length=255)

    def __unicode__(self):
        return self.label


class BlogPost(models.Model):
    """
    Test workflow
    """
    state = FSMField(default='new', protected=True)

    def can_restore(self, user):
        return user.is_superuser or user.is_staff

    @transition(field=state, source='new', target='published',
                on_error='failed', permission='testapp.can_publish_post')
    def publish(self):
        pass

    @transition(field=state, source='published')
    def notify_all(self):
        pass

    @transition(field=state, source='published', target='hidden', on_error='failed',)
    def hide(self):
        pass

    @transition(
        field=state,
        source='new',
        target='removed',
        on_error='failed',
        permission=lambda self, u: u.has_perm('testapp.can_remove_post'))
    def remove(self):
        raise Exception('No rights to delete %s' % self)

    @transition(field=state, source='new', target='restored',
                on_error='failed', permission=can_restore)
    def restore(self):
        pass

    @transition(field=state, source=['published', 'hidden'], target='stolen')
    def steal(self):
        pass

    @transition(field=state, source='*', target='moderated')
    def moderate(self):
        pass

    class Meta:
        permissions = [
            ('can_publish_post', 'Can publish post'),
            ('can_remove_post', 'Can remove post'),
        ]



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, TransitionNotAllowed, \
    transition, can_proceed


def condition_func(instance):
    return True


class BlogPostWithConditions(models.Model):
    state = FSMField(default='new')

    def model_condition(self):
        return True

    def unmet_condition(self):
        return False

    @transition(field=state, source='new', target='published',
                conditions=[condition_func, model_condition])
    def publish(self):
        pass

    @transition(field=state, source='published', target='destroyed',
                conditions=[condition_func, unmet_condition])
    def destroy(self):
        pass


class ConditionalTest(TestCase):
    def setUp(self):
        self.model = BlogPostWithConditions()

    def test_initial_staet(self):
        self.assertEqual(self.model.state, 'new')

    def test_known_transition_should_succeed(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.model.publish()
        self.assertEqual(self.model.state, 'published')

    def test_unmet_condition(self):
        self.model.publish()
        self.assertEqual(self.model.state, 'published')
        self.assertFalse(can_proceed(self.model.destroy))
        self.assertRaises(TransitionNotAllowed, self.model.destroy)

        self.assertTrue(can_proceed(self.model.destroy,
                                    check_conditions=False))



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, ConcurrentTransitionMixin, ConcurrentTransition, transition


class LockedBlogPost(ConcurrentTransitionMixin, models.Model):
    state = FSMField(default='new', protected=True)
    text = models.CharField(max_length=50)

    @transition(field=state, source='new', target='published')
    def publish(self):
        pass

    @transition(field=state, source='published', target='removed')
    def remove(self):
        pass

    class Meta:
        app_label = 'testapp'


class ExtendedBlogPost(LockedBlogPost):
    review_state = FSMField(default='waiting', protected=True)
    notes = models.CharField(max_length=50)

    @transition(field=review_state, source='waiting', target='rejected')
    def reject(self):
        pass

    class Meta:
        app_label = 'testapp'


class TestLockMixin(TestCase):
    def test_create_succeed(self):
        LockedBlogPost.objects.create(text='test_create_succeed')

    def test_crud_succeed(self):
        post = LockedBlogPost(text='test_crud_succeed')
        post.publish()
        post.save()

        post = LockedBlogPost.objects.get(pk=post.pk)
        self.assertEqual('published', post.state)
        post.text = 'test_crud_succeed2'
        post.save()

        post = LockedBlogPost.objects.get(pk=post.pk)
        self.assertEqual('test_crud_succeed2', post.text)

    def test_save_and_change_succeed(self):
        post = LockedBlogPost(text='test_crud_succeed')
        post.publish()
        post.save()

        post.remove()
        post.save()

    def test_concurent_modifications_raise_exception(self):
        post1 = LockedBlogPost.objects.create()
        post2 = LockedBlogPost.objects.get(pk=post1.pk)

        post1.publish()
        post1.save()

        post2.text = 'aaa'
        post2.publish()
        with self.assertRaises(ConcurrentTransition):
            post2.save()

    def test_inheritance_crud_succeed(self):
        post = ExtendedBlogPost(text='test_inheritance_crud_succeed', notes='reject me')
        post.publish()
        post.save()

        post = ExtendedBlogPost.objects.get(pk=post.pk)
        self.assertEqual('published', post.state)
        post.text = 'test_inheritance_crud_succeed2'
        post.reject()
        post.save()

        post = ExtendedBlogPost.objects.get(pk=post.pk)
        self.assertEqual('rejected', post.review_state)
        self.assertEqual('test_inheritance_crud_succeed2', post.text)



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition


class BlogPostWithCustomData(models.Model):
    state = FSMField(default='new')

    @transition(field=state, source='new', target='published', conditions=[],
                custom={'label': 'Publish', 'type': '*'})
    def publish(self):
        pass

    @transition(field=state, source='published', target='destroyed',
                custom=dict(label="Destroy", type='manual'))
    def destroy(self):
        pass

    @transition(field=state, source='published', target='review',
                custom=dict(label="Periodic review", type='automated'))
    def review(self):
        pass

    class Meta:
        app_label = 'testapp'


class CustomTransitionDataTest(TestCase):
    def setUp(self):
        self.model = BlogPostWithCustomData()

    def test_initial_state(self):
        self.assertEqual(self.model.state, 'new')
        transitions = list(self.model.get_available_state_transitions())
        self.assertEquals(len(transitions), 1)
        self.assertEqual(transitions[0].target, 'published')
        self.assertDictEqual(transitions[0].custom, {'label': 'Publish', 'type': '*'})

    def test_all_transitions_have_custom_data(self):
        transitions = self.model.get_all_state_transitions()
        for t in transitions:
            self.assertIsNotNone(t.custom['label'])
            self.assertIsNotNone(t.custom['type'])



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition


class BlogPostWithStringField(models.Model):
    state = FSMField(default='new')

    @transition(field='state', source='new', target='published', conditions=[])
    def publish(self):
        pass

    @transition(field='state', source='published', target='destroyed')
    def destroy(self):
        pass

    @transition(field='state', source='published', target='review')
    def review(self):
        pass

    class Meta:
        app_label = 'testapp'


class StringFieldTestCase(TestCase):
    def setUp(self):
        self.model = BlogPostWithStringField()

    def test_initial_state(self):
        self.assertEqual(self.model.state, 'new')
        self.model.publish()
        self.assertEqual(self.model.state, 'published')



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition
from django_fsm.signals import post_transition


class TestModel(models.Model):
    counter = models.IntegerField(default=0)
    signal_counter = models.IntegerField(default=0)
    state = FSMField(default="SUBMITTED_BY_USER")

    @transition(field=state, source="SUBMITTED_BY_USER", target="REVIEW_USER")
    @transition(field=state, source="SUBMITTED_BY_ADMIN", target="REVIEW_ADMIN")
    @transition(field=state, source="SUBMITTED_BY_ANONYMOUS", target="REVIEW_ANONYMOUS")
    def review(self):
        self.counter += 1

    class Meta:
        app_label = 'testapp'


def count_calls(sender, instance, name, source, target, **kwargs):
    instance.signal_counter += 1


post_transition.connect(count_calls, sender=TestModel)


class TestStateProxy(TestCase):
    def test_transition_method_called_once(self):
        model = TestModel()
        model.review()
        self.assertEqual(1, model.counter)
        self.assertEqual(1, model.signal_counter)



from django.contrib.auth.models import User, Permission
from django.test import TestCase

from django_fsm import has_transition_perm
from testapp.models import BlogPost


class PermissionFSMFieldTest(TestCase):
    def setUp(self):
        self.model = BlogPost()
        self.unpriviledged = User.objects.create(username='unpriviledged')
        self.priviledged = User.objects.create(username='priviledged')
        self.staff = User.objects.create(username='staff', is_staff=True)

        self.priviledged.user_permissions.add(
            Permission.objects.get_by_natural_key('can_publish_post', 'testapp', 'blogpost'))
        self.priviledged.user_permissions.add(
            Permission.objects.get_by_natural_key('can_remove_post', 'testapp', 'blogpost'))

    def test_proviledged_access_succed(self):
        self.assertTrue(has_transition_perm(self.model.publish, self.priviledged))
        self.assertTrue(has_transition_perm(self.model.remove, self.priviledged))

        transitions = self.model.get_available_user_state_transitions(self.priviledged)
        self.assertEquals(set(['publish', 'remove', 'moderate']),
                          set(transition.name for transition in transitions))

    def test_unpriviledged_access_prohibited(self):
        self.assertFalse(has_transition_perm(self.model.publish, self.unpriviledged))
        self.assertFalse(has_transition_perm(self.model.remove, self.unpriviledged))

        transitions = self.model.get_available_user_state_transitions(self.unpriviledged)
        self.assertEquals(set(['moderate']),
                          set(transition.name for transition in transitions))

    def test_permission_instance_method(self):
        self.assertFalse(has_transition_perm(self.model.restore, self.unpriviledged))
        self.assertTrue(has_transition_perm(self.model.restore, self.staff))



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition, RETURN_VALUE, GET_STATE
from django_fsm.signals import pre_transition, post_transition


class MultiResultTest(models.Model):
    state = FSMField(default='new')

    @transition(
        field=state,
        source='new',
        target=RETURN_VALUE('for_moderators', 'published'))
    def publish(self, is_public=False):
        return 'published' if is_public else 'for_moderators'

    @transition(
        field=state,
        source='for_moderators',
        target=GET_STATE(
            lambda self, allowed: 'published' if allowed else 'rejected',
            states=['published', 'rejected']
        )
    )
    def moderate(self, allowed):
        pass

    class Meta:
        app_label = 'testapp'


class Test(TestCase):
    def test_return_state_succeed(self):
        instance = MultiResultTest()
        instance.publish(is_public=True)
        self.assertEqual(instance.state, 'published')

    def test_get_state_succeed(self):
        instance = MultiResultTest(state='for_moderators')
        instance.moderate(allowed=False)
        self.assertEqual(instance.state, 'rejected')


class TestSignals(TestCase):
    def setUp(self):
        self.pre_transition_called = False
        self.post_transition_called = False
        pre_transition.connect(self.on_pre_transition, sender=MultiResultTest)
        post_transition.connect(self.on_post_transition, sender=MultiResultTest)

    def on_pre_transition(self, sender, instance, name, source, target, **kwargs):
        self.assertEqual(instance.state, source)
        self.pre_transition_called = True

    def on_post_transition(self, sender, instance, name, source, target, **kwargs):
        self.assertEqual(instance.state, target)
        self.post_transition_called = True

    def test_signals_called_with_get_state(self):
        instance = MultiResultTest(state='for_moderators')
        instance.moderate(allowed=False)
        self.assertTrue(self.pre_transition_called)
        self.assertTrue(self.post_transition_called)

    def test_signals_called_with_return_value(self):
        instance = MultiResultTest()
        instance.publish(is_public=True)
        self.assertTrue(self.pre_transition_called)
        self.assertTrue(self.post_transition_called)



from django.db import models
from django.test import TestCase

from django_fsm import FSMField, transition, can_proceed


class BaseModel(models.Model):
    state = FSMField(default='new')

    @transition(field=state, source='new', target='published')
    def publish(self):
        pass


class InheritedModel(BaseModel):
    @transition(field='state', source='published', target='sticked')
    def stick(self):
        pass

    class Meta:
        proxy = True


class TestinheritedModel(TestCase):
    def setUp(self):
        self.model = InheritedModel()

    def test_known_transition_should_succeed(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.model.publish()
        self.assertEqual(self.model.state, 'published')

        self.assertTrue(can_proceed(self.model.stick))
        self.model.stick()
        self.assertEqual(self.model.state, 'sticked')

    def test_field_available_transitions_works(self):
        self.model.publish()
        self.assertEqual(self.model.state, 'published')
        transitions = self.model.get_available_state_transitions()
        self.assertEqual(['sticked'], [data.target for data in transitions])

    def test_field_all_transitions_base_model(self):
        transitions = BaseModel().get_all_state_transitions()
        self.assertEqual(set([('new', 'published')]),
                         set((data.source, data.target) for data in transitions))

    def test_field_all_transitions_works(self):
        transitions = self.model.get_all_state_transitions()
        self.assertEqual(set([('new', 'published'),
                              ('published', 'sticked')]),
                         set((data.source, data.target) for data in transitions))



import django

PROJECT_APPS = ('django_fsm', 'testapp',)

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'guardian',
) + PROJECT_APPS

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend', # this is default
    'guardian.backends.ObjectPermissionBackend',
)

DATABASE_ENGINE = 'sqlite3'
SECRET_KEY = 'nokey'
MIDDLEWARE_CLASSES = ()
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

if django.VERSION < (1, 9):
    class DisableMigrations(object):
        def __contains__(self, item):
            return True

        def __getitem__(self, item):
            return 'notmigrations'

    MIGRATION_MODULES = DisableMigrations()
else:
    MIGRATION_MODULES = {
    'auth': None,
    'contenttypes': None,
    'guardian': None,
}


ANONYMOUS_USER_ID = 0



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition, can_proceed


class TestExceptTargetTransitionShortcut(models.Model):
    state = FSMField(default='new')

    @transition(field=state, source='new', target='published')
    def publish(self):
        pass

    @transition(field=state, source='+', target='removed')
    def remove(self):
        pass

    class Meta:
        app_label = 'testapp'


class Test(TestCase):
    def setUp(self):
        self.model = TestExceptTargetTransitionShortcut()

    def test_usecase(self):
        self.assertEqual(self.model.state, 'new')
        self.assertTrue(can_proceed(self.model.remove))
        self.model.remove()

        self.assertEqual(self.model.state, 'removed')
        self.assertFalse(can_proceed(self.model.remove))



try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    # Django 1.6
    from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition


class Ticket(models.Model):

    class Meta:
        app_label = 'testapp'


class Task(models.Model):
    class STATE:
        NEW = 'new'
        DONE = 'done'

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    causality = GenericForeignKey('content_type', 'object_id')
    state = FSMField(default=STATE.NEW)

    @transition(field=state, source=STATE.NEW, target=STATE.DONE)
    def do(self):
        pass

    class Meta:
        app_label = 'testapp'


class Test(TestCase):
    def setUp(self):
        self.ticket = Ticket.objects.create()

    def test_model_objects_create(self):
        """Check a model with state field can be created
        if one of the other fields is a property or a virtual field.
        """
        Task.objects.create(causality=self.ticket)



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition


class WorkflowMixin(object):
    @transition(field='state', source="*", target='draft')
    def draft(self):
        pass

    @transition(field='state', source="draft", target='published')
    def publish(self):
        pass

    class Meta:
        app_label = 'testapp'


class MixinSupportTestModel(WorkflowMixin, models.Model):
    state = FSMField(default="new")


class Test(TestCase):
    def test_usecase(self):
        model = MixinSupportTestModel()

        model.draft()
        self.assertEqual(model.state, 'draft')

        model.publish()
        self.assertEqual(model.state, 'published')



from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition


class Insect(models.Model):
    class STATE:
        CATERPILLAR = 'CTR'
        BUTTERFLY = 'BTF'

    STATE_CHOICES = ((STATE.CATERPILLAR, 'Caterpillar', 'Caterpillar'),
                     (STATE.BUTTERFLY, 'Butterfly', 'Butterfly'))

    state = FSMField(default=STATE.CATERPILLAR, state_choices=STATE_CHOICES)

    @transition(field=state, source=STATE.CATERPILLAR, target=STATE.BUTTERFLY)
    def cocoon(self):
        pass

    def fly(self):
        raise NotImplementedError

    def crawl(self):
        raise NotImplementedError

    class Meta:
        app_label = 'testapp'


class Caterpillar(Insect):
    def crawl(self):
        """
        Do crawl
        """

    class Meta:
        app_label = 'testapp'
        proxy = True


class Butterfly(Insect):
    def fly(self):
        """
        Do fly
        """

    class Meta:
        app_label = 'testapp'
        proxy = True


class TestStateProxy(TestCase):
    def test_initial_proxy_set_succeed(self):
        insect = Insect()
        self.assertTrue(isinstance(insect, Caterpillar))

    def test_transition_proxy_set_succeed(self):
        insect = Insect()
        insect.cocoon()
        self.assertTrue(isinstance(insect, Butterfly))

    def test_load_proxy_set(self):
        Insect.objects.create(state=Insect.STATE.CATERPILLAR)
        Insect.objects.create(state=Insect.STATE.BUTTERFLY)

        insects = Insect.objects.all()
        self.assertEqual(set([Caterpillar, Butterfly]), set(insect.__class__ for insect in insects))



from django.db import models
from django.test import TestCase
from django_fsm import FSMIntegerField, TransitionNotAllowed, transition


class BlogPostStateEnum(object):
    NEW = 10
    PUBLISHED = 20
    HIDDEN = 30


class BlogPostWithIntegerField(models.Model):
    state = FSMIntegerField(default=BlogPostStateEnum.NEW)

    @transition(field=state, source=BlogPostStateEnum.NEW, target=BlogPostStateEnum.PUBLISHED)
    def publish(self):
        pass

    @transition(field=state, source=BlogPostStateEnum.PUBLISHED, target=BlogPostStateEnum.HIDDEN)
    def hide(self):
        pass


class BlogPostWithIntegerFieldTest(TestCase):
    def setUp(self):
        self.model = BlogPostWithIntegerField()

    def test_known_transition_should_succeed(self):
        self.model.publish()
        self.assertEqual(self.model.state, BlogPostStateEnum.PUBLISHED)

        self.model.hide()
        self.assertEqual(self.model.state, BlogPostStateEnum.HIDDEN)

    def test_unknow_transition_fails(self):
        self.assertRaises(TransitionNotAllowed, self.model.hide)



from django.db import models
from django.test import TestCase
from django_fsm import FSMKeyField, TransitionNotAllowed, transition, can_proceed


FK_AVAILABLE_STATES = (
    ('New', '_NEW_'),
    ('Published', '_PUBLISHED_'),
    ('Hidden', '_HIDDEN_'),
    ('Removed', '_REMOVED_'),
    ('Stolen', '_STOLEN_'),
    ('Moderated', '_MODERATED_'))


class DBState(models.Model):
    id = models.CharField(primary_key=True, max_length=50)

    label = models.CharField(max_length=255)

    def __unicode__(self):
        return self.label

    class Meta:
        app_label = 'django_fsm'


class FKBlogPost(models.Model):
    state = FSMKeyField(DBState, default='new', protected=True, on_delete=models.CASCADE)

    @transition(field=state, source='new', target='published')
    def publish(self):
        pass

    @transition(field=state, source='published')
    def notify_all(self):
        pass

    @transition(field=state, source='published', target='hidden')
    def hide(self):
        pass

    @transition(field=state, source='new', target='removed')
    def remove(self):
        raise Exception('Upss')

    @transition(field=state, source=['published', 'hidden'], target='stolen')
    def steal(self):
        pass

    @transition(field=state, source='*', target='moderated')
    def moderate(self):
        pass

    class Meta:
        app_label = 'django_fsm'


class FSMKeyFieldTest(TestCase):
    def setUp(self):
        for item in FK_AVAILABLE_STATES:
            DBState.objects.create(pk=item[0], label=item[1])
        self.model = FKBlogPost()

    def test_initial_state_instatiated(self):
        self.assertEqual(self.model.state, 'new',)

    def test_known_transition_should_succeed(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.model.publish()
        self.assertEqual(self.model.state, 'published')

        self.assertTrue(can_proceed(self.model.hide))
        self.model.hide()
        self.assertEqual(self.model.state, 'hidden')

    def test_unknow_transition_fails(self):
        self.assertFalse(can_proceed(self.model.hide))
        self.assertRaises(TransitionNotAllowed, self.model.hide)

    def test_state_non_changed_after_fail(self):
        self.assertTrue(can_proceed(self.model.remove))
        self.assertRaises(Exception, self.model.remove)
        self.assertEqual(self.model.state, 'new')

    def test_allowed_null_transition_should_succeed(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.model.publish()
        self.model.notify_all()
        self.assertEqual(self.model.state, 'published')

    def test_unknow_null_transition_should_fail(self):
        self.assertRaises(TransitionNotAllowed, self.model.notify_all)
        self.assertEqual(self.model.state, 'new')

    def test_mutiple_source_support_path_1_works(self):
        self.model.publish()
        self.model.steal()
        self.assertEqual(self.model.state, 'stolen')

    def test_mutiple_source_support_path_2_works(self):
        self.model.publish()
        self.model.hide()
        self.model.steal()
        self.assertEqual(self.model.state, 'stolen')

    def test_star_shortcut_succeed(self):
        self.assertTrue(can_proceed(self.model.moderate))
        self.model.moderate()
        self.assertEqual(self.model.state, 'moderated')


"""
TODO FIX it
class BlogPostStatus(models.Model):
    name = models.CharField(max_length=10, unique=True)
    objects = models.Manager()

    class Meta:
        app_label = 'django_fsm'


class BlogPostWithFKState(models.Model):
    status = FSMKeyField(BlogPostStatus, default=lambda: BlogPostStatus.objects.get(name="new"))

    @transition(field=status, source='new', target='published')
    def publish(self):
        pass

    @transition(field=status, source='published', target='hidden')
    def hide(self):
        pass


class BlogPostWithFKStateTest(TestCase):
    def setUp(self):
        BlogPostStatus.objects.create(name="new")
        BlogPostStatus.objects.create(name="published")
        BlogPostStatus.objects.create(name="hidden")
        self.model = BlogPostWithFKState()

    def test_known_transition_should_succeed(self):
        self.model.publish()
        self.assertEqual(self.model.state, 'published')

        self.model.hide()
        self.assertEqual(self.model.state, 'hidden')

    def test_unknow_transition_fails(self):
        self.assertRaises(TransitionNotAllowed, self.model.hide)
"""



from django.db import models
from django.test import TestCase

from django_fsm import FSMField, transition, can_proceed
from django_fsm.signals import post_transition


class ExceptionalBlogPost(models.Model):
    state = FSMField(default='new')

    @transition(field=state, source='new', target='published', on_error='crashed')
    def publish(self):
        raise Exception('Upss')

    @transition(field=state, source='new', target='deleted')
    def delete(self):
        raise Exception('Upss')

    class Meta:
        app_label = 'testapp'


class FSMFieldExceptionTest(TestCase):
    def setUp(self):
        self.model = ExceptionalBlogPost()
        post_transition.connect(self.on_post_transition, sender=ExceptionalBlogPost)
        self.post_transition_data = None

    def on_post_transition(self, **kwargs):
        self.post_transition_data = kwargs

    def test_state_changed_after_fail(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.assertRaises(Exception, self.model.publish)
        self.assertEqual(self.model.state, 'crashed')
        self.assertEqual(self.post_transition_data['target'], 'crashed')
        self.assertTrue('exception' in self.post_transition_data)

    def test_state_not_changed_after_fail(self):
        self.assertTrue(can_proceed(self.model.delete))
        self.assertRaises(Exception, self.model.delete)
        self.assertEqual(self.model.state, 'new')
        self.assertIsNone(self.post_transition_data)



from django.db import models
from django.test import TestCase

from django_fsm import FSMField, TransitionNotAllowed, transition, can_proceed
from django_fsm.signals import pre_transition, post_transition


class BlogPost(models.Model):
    state = FSMField(default='new')

    @transition(field=state, source='new', target='published')
    def publish(self):
        pass

    @transition(source='published', field=state)
    def notify_all(self):
        pass

    @transition(source='published', target='hidden', field=state)
    def hide(self):
        pass

    @transition(source='new', target='removed', field=state)
    def remove(self):
        raise Exception('Upss')

    @transition(source=['published', 'hidden'], target='stolen', field=state)
    def steal(self):
        pass

    @transition(source='*', target='moderated', field=state)
    def moderate(self):
        pass

    @transition(source='+', target='blocked', field=state)
    def block(self):
        pass

    @transition(source='*', target='', field=state)
    def empty(self):
        pass


class FSMFieldTest(TestCase):
    def setUp(self):
        self.model = BlogPost()

    def test_initial_state_instantiated(self):
        self.assertEqual(self.model.state, 'new')

    def test_known_transition_should_succeed(self):
        self.assertTrue(can_proceed(self.model.publish))
        self.model.publish()
        self.assertEqual(self.model.state, 'published')

        self.assertTrue(can_proceed(self.model.hide))
        self.model.hide()
        self.assertEqual(self.model.state, 'hidden')

    def test_unknown_transition_fails(self):
        self.assertFalse(can_proceed(self.model.hide))
        self.assertRaises(TransitionNotAllowed, self.model.hide)

    def test_state_non_changed_after_fail(self):
        self.assertTrue(can_proceed(self.model.remove))
        self.assertRaises(Exception, self.model.remove)
        self.assertEqual(self.model.state, 'new')

    def test_allowed_null_transition_should_succeed(self):
        self.model.publish()
        self.model.notify_all()
        self.assertEqual(self.model.state, 'published')

    def test_unknown_null_transition_should_fail(self):
        self.assertRaises(TransitionNotAllowed, self.model.notify_all)
        self.assertEqual(self.model.state, 'new')

    def test_multiple_source_support_path_1_works(self):
        self.model.publish()
        self.model.steal()
        self.assertEqual(self.model.state, 'stolen')

    def test_multiple_source_support_path_2_works(self):
        self.model.publish()
        self.model.hide()
        self.model.steal()
        self.assertEqual(self.model.state, 'stolen')

    def test_star_shortcut_succeed(self):
        self.assertTrue(can_proceed(self.model.moderate))
        self.model.moderate()
        self.assertEqual(self.model.state, 'moderated')

    def test_plus_shortcut_succeeds_for_other_source(self):
        """Tests that the '+' shortcut succeeds for a source
        other than the target.
        """
        self.assertTrue(can_proceed(self.model.block))
        self.model.block()
        self.assertEqual(self.model.state, 'blocked')

    def test_plus_shortcut_fails_for_same_source(self):
        """Tests that the '+' shortcut fails if the source
        equals the target.
        """
        self.model.block()
        self.assertFalse(can_proceed(self.model.block))
        self.assertRaises(TransitionNotAllowed, self.model.block)

    def test_empty_string_target(self):
        self.model.empty()
        self.assertEqual(self.model.state, '')


class StateSignalsTests(TestCase):
    def setUp(self):
        self.model = BlogPost()
        self.pre_transition_called = False
        self.post_transition_called = False
        pre_transition.connect(self.on_pre_transition, sender=BlogPost)
        post_transition.connect(self.on_post_transition, sender=BlogPost)

    def on_pre_transition(self, sender, instance, name, source, target, **kwargs):
        self.assertEqual(instance.state, source)
        self.pre_transition_called = True

    def on_post_transition(self, sender, instance, name, source, target, **kwargs):
        self.assertEqual(instance.state, target)
        self.post_transition_called = True

    def test_signals_called_on_valid_transition(self):
        self.model.publish()
        self.assertTrue(self.pre_transition_called)
        self.assertTrue(self.post_transition_called)

    def test_signals_not_called_on_invalid_transition(self):
        self.assertRaises(TransitionNotAllowed, self.model.hide)
        self.assertFalse(self.pre_transition_called)
        self.assertFalse(self.post_transition_called)


class TestFieldTransitionsInspect(TestCase):
    def setUp(self):
        self.model = BlogPost()

    def test_available_conditions_from_new(self):
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('new', 'published'),
                        ('new', 'removed'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)

    def test_available_conditions_from_published(self):
        self.model.publish()
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('published', None),
                        ('published', 'hidden'),
                        ('published', 'stolen'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)

    def test_available_conditions_from_hidden(self):
        self.model.publish()
        self.model.hide()
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('hidden', 'stolen'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)

    def test_available_conditions_from_stolen(self):
        self.model.publish()
        self.model.steal()
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)

    def test_available_conditions_from_blocked(self):
        self.model.block()
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('*', '')])
        self.assertEqual(actual, expected)

    def test_available_conditions_from_empty(self):
        self.model.empty()
        transitions = self.model.get_available_state_transitions()
        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)

    def test_all_conditions(self):
        transitions = self.model.get_all_state_transitions()

        actual = set((transition.source, transition.target) for transition in transitions)
        expected = set([('*', 'moderated'),
                        ('new', 'published'),
                        ('new', 'removed'),
                        ('published', None),
                        ('published', 'hidden'),
                        ('published', 'stolen'),
                        ('hidden', 'stolen'),
                        ('*', ''),
                        ('+', 'blocked')])
        self.assertEqual(actual, expected)



from django.db import models
from django.test import TestCase

from django_fsm import FSMField, transition


class ProtectedAccessModel(models.Model):
    status = FSMField(default='new', protected=True)

    @transition(field=status, source='new', target='published')
    def publish(self):
        pass

    class Meta:
        app_label = 'django_fsm'


class TestDirectAccessModels(TestCase):
    def test_no_direct_access(self):
        instance = ProtectedAccessModel()
        self.assertEqual(instance.status, 'new')

        def try_change():
            instance.status = 'change'

        self.assertRaises(AttributeError, try_change)

        instance.publish()
        instance.save()
        self.assertEqual(instance.status, 'published')
